import datetime

import pytest

from stechuhr import logbooks as lgbks

import hypothesis as hyp
from hypothesis import strategies as st

from stechuhr.errors import InAfterOutError


def test__in_1970_1_1_0_0_0_out_none() -> None:
    in_datetimestamp = datetime.datetime(1970, 1, 1, 0, 0, 0)
    out_datetimestamp = None
    expected = lgbks.Entry(
        in_datetimestamp=in_datetimestamp, out_datetimestamp=out_datetimestamp
    )

    actual = lgbks.Entry.from_datetimestamps(
        in_datetimestamp=in_datetimestamp, out_datetimestamp=out_datetimestamp
    )

    assert actual == expected


def test__in_1970_1_1_0_0_0_out_2000_12_31_12_31_59() -> None:
    in_datetimestamp = datetime.datetime(1970, 1, 1, 0, 0, 0)
    out_datetimestamp = datetime.datetime(2000, 12, 31, 12, 31, 59)
    expected = lgbks.Entry(
        in_datetimestamp=in_datetimestamp, out_datetimestamp=out_datetimestamp
    )

    actual = lgbks.Entry.from_datetimestamps(
        in_datetimestamp=in_datetimestamp, out_datetimestamp=out_datetimestamp
    )

    assert actual == expected


@hyp.given(st.datetimes(), st.datetimes())
def test__in_before_out(fst_datetimestamp, snd_datetimestamp) -> None:
    in_datetimestamp, out_datetimestamp = sorted((fst_datetimestamp, snd_datetimestamp))
    expected = lgbks.Entry(
        in_datetimestamp=in_datetimestamp, out_datetimestamp=out_datetimestamp
    )

    actual = lgbks.Entry.from_datetimestamps(
        in_datetimestamp=in_datetimestamp, out_datetimestamp=out_datetimestamp
    )

    assert actual == expected


def test__in__2000_12_31_12_31_59_out_1970_1_1_0_0_0__raises_in_after_out_error() -> (
    None
):
    in_datetimestamp = datetime.datetime(2000, 12, 31, 12, 31, 59)
    out_datetimestamp = datetime.datetime(1970, 1, 1, 0, 0, 0)

    with pytest.raises(InAfterOutError):
        lgbks.Entry.from_datetimestamps(
            in_datetimestamp=in_datetimestamp, out_datetimestamp=out_datetimestamp
        )


@hyp.given(st.datetimes(), st.datetimes())
def test__in_after_out__raises_in_after_out_error(
    fst_datetimestamp, snd_datetimestamp
) -> None:
    hyp.assume(fst_datetimestamp != snd_datetimestamp)
    in_datetimestamp, out_datetimestamp = sorted(
        (fst_datetimestamp, snd_datetimestamp), reverse=True
    )

    with pytest.raises(InAfterOutError):
        lgbks.Entry.from_datetimestamps(
            in_datetimestamp=in_datetimestamp, out_datetimestamp=out_datetimestamp
        )



import datetime

from typing import TYPE_CHECKING


class InAfterOutError(ValueError):
    def __init__(
        self, in_datetimestamp: datetime.datetime, out_datetimestamp: datetime.datetime
    ) -> None:
        super().__init__(f"{in_datetimestamp=} > {out_datetimestamp=}")


class OverlappingEntriesError(ValueError):
    def __init__(self, entry1: "Entry", entry2: "Entry") -> None:
        super().__init__(f"{entry1=} overlaps with {entry2=}")


if TYPE_CHECKING:
    from stechuhr.logbooks import Entry



import datetime
from collections.abc import Iterable
from enum import Enum, auto

from hypothesis import strategies as st

from stechuhr import logbooks as lgbks


class EntryState(Enum):
    CLOSED = (auto(),)
    OPEN = (auto(),)
    UNDETERMINED = (auto(),)


@st.composite
def entries(
    draw,
    *,
    min_value: datetime.datetime = datetime.datetime.min,
    max_value: datetime.datetime = datetime.datetime.max,
    unfinished: EntryState = EntryState.UNDETERMINED,
) -> lgbks.Entry:
    in_datetimestamp = draw(st.datetimes(min_value=min_value, max_value=max_value))
    out_datetimestamp = None
    if unfinished is EntryState.UNDETERMINED:
        unfinished = draw(st.sampled_from((EntryState.CLOSED, EntryState.OPEN)))
    match unfinished:
        case EntryState.CLOSED:
            out_datetimestamp = draw(
                st.datetimes(min_value=in_datetimestamp, max_value=max_value)
            )
        case EntryState.OPEN:
            out_datetimestamp = None
        case EntryState.UNDETERMINED:
            assert False
    return lgbks.Entry.from_datetimestamps(in_datetimestamp, out_datetimestamp)


@st.composite
def non_overlapping_entries(
    draw,
    *,
    min_size: int = 0,
    max_size: int | None = None,
    min_value: datetime.datetime = datetime.datetime.min,
    max_value: datetime.datetime = datetime.datetime.max,
    last_unfinished: EntryState = EntryState.UNDETERMINED,
) -> Iterable[lgbks.Entry]:
    if max_size is None:
        max_size = draw(st.integers(min_value=min_size))
    target_size = draw(st.integers(min_value=min_size, max_value=max_size))
    if target_size == 0:
        return
    entry = draw(
        entries(
            min_value=min_value,
            max_value=max_value,
            unfinished=EntryState.CLOSED if target_size > 1 else last_unfinished,
        )
    )
    assert target_size <= 1 or entry.out_datetimestamp is not None, entry
    yield entry
    target_size -= 1
    while target_size > 0:
        assert entry.out_datetimestamp is not None, entry
        entry = draw(
            entries(
                min_value=entry.out_datetimestamp,
                max_value=max_value,
                unfinished=EntryState.CLOSED if target_size > 1 else last_unfinished,
            )
        )
        assert target_size <= 1 or entry.out_datetimestamp is not None, entry
        yield entry
        target_size -= 1


@st.composite
def overlapping_entries(
    draw,
    *,
    min_size: int = 2,
    max_size: int | None = None,
    min_value: datetime.datetime = datetime.datetime.min,
    max_value: datetime.datetime = datetime.datetime.max,
    last_unfinished: EntryState = EntryState.UNDETERMINED,
) -> Iterable[lgbks.Entry]:
    if min_size < 2:
        raise ValueError(
            f"Cannot produce overlapping entries with less than 2 entries {min_size=}"
        )
    if max_size is None:
        max_size = draw(st.integers(min_value=min_size))
    assert max_size is not None

    entries_ = draw(
        st.lists(
            entries(unfinished=EntryState.CLOSED),
            min_size=min_size - 1,
            max_size=max_size - 1,
        )
    )
    entry = draw(st.sampled_from(entries_))
    assert entry is not None
    raise NotImplementedError



import datetime
from collections.abc import Iterable
from dataclasses import dataclass
from typing import Self

from stechuhr.errors import InAfterOutError


@dataclass(frozen=True, order=True)
class Entry:
    in_datetimestamp: datetime.datetime
    out_datetimestamp: datetime.datetime | None

    @classmethod
    def from_datetimestamps(
        cls,
        in_datetimestamp: datetime.datetime,
        out_datetimestamp: datetime.datetime | None = None,
    ) -> Self:
        if out_datetimestamp is not None and in_datetimestamp > out_datetimestamp:
            raise InAfterOutError(
                in_datetimestamp=in_datetimestamp, out_datetimestamp=out_datetimestamp
            )
        return cls(
            in_datetimestamp=in_datetimestamp, out_datetimestamp=out_datetimestamp
        )


@dataclass(frozen=True)
class Logbook:
    entries: frozenset[Entry]

    @classmethod
    def from_entries(cls, entries: Iterable[Entry]) -> Self:
        return cls(frozenset(entries))
